import { CallableObject, isSetEqual, Debouncer, isDefined } from '@s-libs/js-core';
import { isUndefined, every, clone, omit, keys } from '@s-libs/micro-dash';
import { Observable, ReplaySubject } from 'rxjs';
import { distinctUntilChanged, map, filter } from 'rxjs/operators';
import { mapAndCacheObjectElements } from '@s-libs/rxjs-core';

class Store extends CallableObject {
    constructor(getRootStore) {
        super((childKey) => {
            var _a, _b, _c;
            return (_c = (_b = (_a = this.activeChildren.get(childKey)) === null || _a === void 0 ? void 0 : _a.values().next()) === null || _b === void 0 ? void 0 : _b.value) !== null && _c !== void 0 ? _c : new ChildStore(getRootStore, this, childKey);
        });
        this.getRootStore = getRootStore;
        /**
         * An `Observable` of the state of this store object.
         */
        this.$ = new Observable((subscriber) => {
            const value = this.state();
            this.subscribers.set(subscriber, value);
            this.maybeActivate();
            subscriber.next(value);
            return () => {
                this.subscribers.delete(subscriber);
                this.maybeDeactivate();
            };
        });
        this.subscribers = new Map();
        this.activeChildren = new Map();
        this.lastKnownStateChanged = false;
    }
    /**
     * Assigns the given values to state of this store object. The resulting state will be like `Object.assign(store.state(), value)`.
     */
    assign(value) {
        this.setUsing((state) => {
            if (isUndefined(state)) {
                throw new Error('cannot assign to undefined state');
            }
            if (every(value, (innerValue, key) => state[key] === innerValue)) {
                return state;
            }
            else {
                return Object.assign(Object.assign({}, state), value);
            }
        });
    }
    /**
     * Runs `func` on the state and replaces it with the return value. The first argument to `func` will be the state, followed by the arguments in `args`.
     *
     * WARNING: You SHOULD NOT use a function that will mutate the state.
     */
    setUsing(func, ...args) {
        this.set(func(this.state(), ...args));
    }
    /**
     * Runs `func` on a shallow clone of the state, replacing the state with the clone. The first argument to `func` will be the cloned state, followed by the arguments in `args`.
     *
     * WARNING: You SHOULD NOT use a function that will mutate nested objects within the state.
     */
    mutateUsing(func, ...args) {
        const state = clone(this.state());
        func(state, ...args);
        this.set(state);
    }
    updateState(value) {
        if (value === this.lastKnownState) {
            return;
        }
        this.lastKnownState = value;
        this.lastKnownStateChanged = true;
        this.activeChildren.forEach((children, key) => {
            for (const child of children) {
                child.updateState(value === null || value === void 0 ? void 0 : value[key]);
            }
        });
    }
    maybeEmit() {
        if (!this.lastKnownStateChanged) {
            return;
        }
        this.lastKnownStateChanged = false;
        this.subscribers.forEach((lastEmitted, subscriber) => {
            if (lastEmitted !== this.lastKnownState) {
                subscriber.next(this.lastKnownState);
                this.subscribers.set(subscriber, this.lastKnownState);
            }
        });
        this.activeChildren.forEach((children) => {
            // eslint-disable-next-line @typescript-eslint/no-unnecessary-condition -- `children` can be undefined if emitting from a previous key removed all subscribers to this key
            for (const child of children !== null && children !== void 0 ? children : []) {
                child.maybeEmit();
            }
        });
    }
    isChildActive(parent, key, child) {
        var _a, _b;
        return (_b = (_a = parent.activeChildren.get(key)) === null || _a === void 0 ? void 0 : _a.has(child)) !== null && _b !== void 0 ? _b : false;
    }
    activateChild(parent, key, child) {
        let set = parent.activeChildren.get(key);
        if (!set) {
            set = new Set();
            parent.activeChildren.set(key, set);
        }
        set.add(child);
        parent.maybeActivate();
    }
    deactivateChild(parent, key, child) {
        const set = parent.activeChildren.get(key);
        set.delete(child);
        if (set.size === 0) {
            parent.activeChildren.delete(key);
            parent.maybeDeactivate();
        }
    }
}

class RootStore extends Store {
    constructor(state) {
        super(() => this);
        this.batchCount = 0;
        this.set(state);
    }
    set(value) {
        this.updateState(value);
        this.maybeEmit();
    }
    delete() {
        this.set(undefined);
    }
    state() {
        return this.lastKnownState;
    }
    /**
     * Turns off this store's observables while `func` is executing, emitting only once afterward, if the store changed. This allows you to batch multiple mutations into a single update at the end.
     *
     * ```ts
     * store.batch(() => {
     *   store.assign({ key1: value1 });
     *   store('key2').delete();
     *   store('key3').set({ key4: value4 });
     *
     *   store('key1').state(); // returns `value1`
     * });
     * ```
     */
    batch(func) {
        ++this.batchCount;
        try {
            func();
        }
        finally {
            if (--this.batchCount === 0) {
                this.maybeEmit();
            }
        }
    }
    refersToSameStateAs(other) {
        return this === other;
    }
    maybeEmit() {
        if (this.batchCount === 0) {
            super.maybeEmit();
        }
    }
    maybeActivate() {
        // root stores are always active
    }
    maybeDeactivate() {
        // root stores are always active
    }
}

class ChildStore extends Store {
    constructor(getRootStore, parent, key) {
        super(getRootStore);
        this.parent = parent;
        this.key = key;
    }
    set(value) {
        if (value === this.state()) {
            return;
        }
        const parentState = clone(this.parent.state());
        if (isUndefined(parentState)) {
            throw new Error('cannot modify when parent state is missing');
        }
        parentState[this.key] = value;
        this.parent.set(parentState);
    }
    delete() {
        this.parent.setUsing(omit, this.key);
    }
    state() {
        var _a;
        if (this.isActive()) {
            return this.lastKnownState;
        }
        else {
            return (_a = this.parent.state()) === null || _a === void 0 ? void 0 : _a[this.key];
        }
    }
    refersToSameStateAs(other) {
        return (other instanceof ChildStore &&
            this.key === other.key &&
            this.parent.refersToSameStateAs(other.parent));
    }
    maybeActivate() {
        var _a;
        if (!this.isActive() && this.shouldBeActive()) {
            this.activateChild(this.parent, this.key, this);
            this.lastKnownState = (_a = this.parent.state()) === null || _a === void 0 ? void 0 : _a[this.key];
        }
    }
    maybeDeactivate() {
        if (this.isActive() && !this.shouldBeActive()) {
            this.deactivateChild(this.parent, this.key, this);
        }
    }
    shouldBeActive() {
        return this.subscribers.size > 0 || this.activeChildren.size > 0;
    }
    isActive() {
        return this.isChildActive(this.parent, this.key, this);
    }
}

// this exists to resolve the order of some circular imports

/**
 * Adds `item` to the end of the array in `store`. Returns a store object representing the newly added item.
 *
 * ```ts
 * declare const store: Store<number[]>; // assume store starts with [1, 2]
 * const added = pushToStoreArray(store, 3); // store has [1, 2, 3]
 * added.set(17); // store has [1, 2, 17]
 * ```
 */
function pushToStoreArray(store, item) {
    const itemStore = store(store.state().length);
    itemStore.set(item);
    return itemStore;
}

/**
 * Returns an observable that emits an array of store objects, one for each element in `source`'s state. The resulting arrays will have references to the exact store objects included in the previous emission when possible, making them performant to use in `*ngFor` expressions without the need to use `trackBy`.
 *
 * ```ts
 * @Component({
 *   template: `
 *     <hero
 *       *ngFor="let heroStore of heroStores$ | async"
 *       [heroStore]="heroStore"
 *     ></hero>
 *   `,
 * })
 * class HeroListComponent {
 *   heroStores$: Observable<Array<Store<Hero>>>;
 *   @Input() private heroesStore: Store<Array<Hero>>;
 *
 *   ngOnChanges() {
 *     this.heroStores$ = spreadArrayStore(this.heroesStore);
 *   }
 * }
 * ```
 */
function spreadArrayStore$(source) {
    let cache = [];
    return source.$.pipe(distinctUntilChanged((x, y) => { var _a, _b; return ((_a = x === null || x === void 0 ? void 0 : x.length) !== null && _a !== void 0 ? _a : 0) === ((_b = y === null || y === void 0 ? void 0 : y.length) !== null && _b !== void 0 ? _b : 0); }), map((array) => {
        array !== null && array !== void 0 ? array : (array = []);
        if (array.length < cache.length) {
            cache = cache.slice(0, array.length);
        }
        else {
            cache = cache.slice();
            for (let i = cache.length; i < array.length; ++i) {
                cache[i] = source(i);
            }
        }
        return cache;
    }));
}

/**
 * Returns an observable that emits an array of store objects, one for each element in `source`'s state. The resulting arrays will have references to the exact store objects included in the previous emission when possible, making them performant to use in `*ngFor` expressions without the need to use `trackBy`.
 *
 * ```ts
 * interface HeroMap {
 *   [name: string]: Hero;
 * }
 *
 * @Component({
 *   template: `
 *     <hero
 *       *ngFor="let heroStore of heroStores$ | async"
 *       [heroStore]="heroStore"
 *     ></hero>
 *   `,
 * })
 * class HeroListComponent {
 *   heroStores$: Observable<Array<Store<Hero>>>;
 *   @Input() private heroesStore: Store<HeroMap>;
 *
 *   ngOnChanges() {
 *     this.heroStores$ = spreadObjectStore(this.heroesStore);
 *   }
 * }
 * ```
 */
function spreadObjectStore$(source) {
    let lastKeySet;
    return source.$.pipe(filter((value) => {
        const keySet = new Set(keys(value));
        if (lastKeySet && isSetEqual(keySet, lastKeySet)) {
            return false;
        }
        lastKeySet = keySet;
        return true;
    }), mapAndCacheObjectElements((_value, key) => key, (_value, key) => source(key)));
}

class UndoManager {
    /**
     * @param maxDepth The maximum size of the history before discarding the oldest state. `0` means no limit.
     */
    constructor(store, maxDepth = 0) {
        this.store = store;
        this.maxDepth = maxDepth;
        this.stack = [];
        this.canUndoSubject = new ReplaySubject(1);
        this.canRedoSubject = new ReplaySubject(1);
        this.stateSubject = new ReplaySubject(1);
        this.collectDebouncer = new Debouncer();
        /**
         * An observable that emits the result of `canUndo()` every time that value changes.
         */
        this.canUndo$ = this.canUndoSubject.pipe(distinctUntilChanged());
        /**
         * An observable that emits the result of `canRedo()` every time that value changes.
         */
        this.canRedo$ = this.canRedoSubject.pipe(distinctUntilChanged());
        /**
         * An observable that emits the current state every time it changes.
         */
        this.state$ = this.stateSubject.pipe(distinctUntilChanged());
        this.reset();
    }
    /**
     * Discard all history and push the current state.
     */
    reset() {
        this.currentStateIndex = -1;
        this.pushCurrentState();
    }
    /**
     * Add the current state to the undo history. Any states that could be reached using `redo()` are discarded.
     *
     * @param collectKey When specified, multiple pushes in a row with the same key will be collected into a single undo state. This is useful e.g. for collecting changes from a text input into larger undo states, rather than undoing one character at a time. After a push with a different key or another undo change (like `.undo()` or `.reset()`), collecting will stop and the next push will be in a new undo state.
     * @param collectDebounce If at least this many milliseconds elapse with no other push, the next one will be to a new undo state regardless of its `collectKey`. Defaults to `undefined`, which sets no such timeout.
     */
    pushCurrentState({ collectKey = undefined, collectDebounce = undefined, } = {}) {
        const nextState = this.extractUndoState(this.store.state());
        if (this.currentStateIndex >= 0 && !this.shouldPush(nextState)) {
            return;
        }
        if (this.prepush(nextState, collectKey)) {
            this.actuallyPush(nextState, collectKey, collectDebounce);
        }
    }
    /**
     * @returns whether any states are available for `undo()`
     */
    canUndo() {
        return this.currentStateIndex > 0;
    }
    /**
     * @returns whether any states are available for `redo()`
     */
    canRedo() {
        return this.currentStateIndex < this.stack.length - 1;
    }
    /**
     * Move backward one step in the history of states saved via `pushCurrentState()`, setting the store to contain that state again.
     *
     * @throws Error when there is no such state (i.e. when `canUndo()` returns false)
     */
    undo() {
        if (!this.canUndo()) {
            throw new Error('Cannot undo');
        }
        this.changeState(-1, 'undo');
    }
    /**
     * Move forward one step in the history of states saved via `pushCurrentState()`, setting the store to contain that state again.
     *
     * @throws Error when there is no such state (i.e. when `canRedo()` returns false)
     */
    redo() {
        if (!this.canRedo()) {
            throw new Error('Cannot redo');
        }
        this.changeState(1, 'redo');
    }
    /**
     * Drops the state from the internal undo stack that would be applied by a call to `.undo()`. This is useful e.g. if making multiple changes in a row that should be collapsed into a single undo state: call this first before pushing the new version.
     *
     * @throws Error when there is no such state (i.e. when `canUndo()` returns false)
     */
    dropCurrentUndoState() {
        if (!this.canUndo()) {
            throw new Error('Nothing to drop');
        }
        --this.currentStateIndex;
        this.dropRedoHistory();
        this.emitUndoChanges();
    }
    /**
     * Returns the current undo state that was most recently pushed or applied. Calls to `undo` will apply the state before this in teh stack, and `redo` will apply the state after this.
     */
    get currentUndoState() {
        return this.stack[this.currentStateIndex];
    }
    /**
     * Returns a view of the internal undo stack, from oldest to newest. Note that this contains states that would be applied by calls to both `.undo()` and `.redo`.
     */
    get undoStack() {
        return this.stack.slice();
    }
    /**
     * Used to determine whether `.pushCurrentState()` actually does anything. Override this e.g. to prevent pushing a duplicate undo state using something like this:
     *
     * ```ts
     * protected shouldPush(state: UndoStateType) {
     *   return equal(state, this.currentUndoState);
     * }
     * ```
     */
    shouldPush(_state) {
        return true;
    }
    dropRedoHistory() {
        this.stack.splice(this.currentStateIndex + 1, this.stack.length);
    }
    /**
     * Each time a state is added to the history, this method will be called to determine whether the oldest state should be dropped. Override to implement more complex logic than the simple `maxDepth`.
     */
    isOverSize(size) {
        return this.maxDepth > 0 && size > this.maxDepth;
    }
    changeState(change, undoOrRedo) {
        const stateToOverwrite = this.currentUndoState;
        this.currentStateIndex += change;
        const stateToApply = this.currentUndoState;
        this.store.getRootStore().batch(() => {
            this.applyUndoState(stateToApply, undoOrRedo, stateToOverwrite);
        });
        this.emitUndoChanges();
    }
    emitUndoChanges(collectKey, collectDebounce) {
        this.canUndoSubject.next(this.canUndo());
        this.canRedoSubject.next(this.canRedo());
        this.stateSubject.next(this.stack[this.currentStateIndex]);
        this.currentCollectKey = collectKey;
        if (collectDebounce !== undefined) {
            this.collectDebouncer.run(() => {
                this.currentCollectKey = undefined;
            }, collectDebounce);
        }
    }
    prepush(nextState, collectKey) {
        if (isDefined(collectKey) && this.currentCollectKey === collectKey) {
            --this.currentStateIndex;
            this.dropRedoHistory();
            if (!this.shouldPush(nextState)) {
                return false;
            }
        }
        return true;
    }
    actuallyPush(nextState, collectKey, collectDebounce) {
        ++this.currentStateIndex;
        this.stack[this.currentStateIndex] = nextState;
        this.dropRedoHistory();
        while (this.stack.length > 1 && this.isOverSize(this.stack.length)) {
            this.stack.shift();
            --this.currentStateIndex;
        }
        this.emitUndoChanges(collectKey, collectDebounce);
    }
}

/*
 * Public API Surface of app-state
 */

/**
 * Generated bundle index. Do not edit.
 */

export { RootStore, Store, UndoManager, pushToStoreArray, spreadArrayStore$, spreadObjectStore$ };
